import * as ts from "typescript";
/**
 * ==============================
 * File Type Utilities
 * ==============================
 */
/**
 * Checks if the provided file name corresponds to a supported file type (TypeScript or JavaScript).
 * @param fileName - The name of the file to check.
 * @returns True if the file is a .ts, .tsx, .js, or .jsx file; false otherwise.
 */
export declare function isSupportedFileType(fileName: string): boolean;
/**
 * ==============================
 * AST Node Utilities
 * ==============================
 */
/**
 * Finds the AST node at the specified position within the source file.
 * @param sourceFile - The source file to search.
 * @param position - The position within the file.
 * @returns The node at the specified position, or undefined if not found.
 */
export declare function findNodeAtPosition(sourceFile: ts.SourceFile, position: number): ts.Node | undefined;
/**
 * Retrieves all preceding comments (both JSDoc and regular comments) from a given AST node.
 * @param node - The AST node.
 * @returns An array of comment strings preceding the node.
 */
export declare function getCommentsFromNode(node: ts.Node): string[];
/**
 * Checks if an AST node is marked as deprecated.
 * @param checker - The TypeScript type checker.
 * @param node - The AST node.
 * @returns True if the node or its symbol is deprecated; false otherwise.
 */
export declare function isNodeDeprecated(checker: ts.TypeChecker, node: ts.Node): boolean;
/**
 * ==============================
 * Symbol Utilities
 * ==============================
 */
/**
 * Retrieves the symbol associated with a given AST node.
 * @param checker - The TypeScript type checker.
 * @param node - The AST node.
 * @returns The symbol associated with the node, or undefined if not found.
 */
export declare function getSymbolAtNode(checker: ts.TypeChecker, node: ts.Node): ts.Symbol | undefined;
/**
 * Retrieves all preceding comments from all declarations of a given symbol.
 * @param symbol - The TypeScript symbol.
 * @returns An array of comment strings from all declarations of the symbol.
 */
export declare function getCommentsFromSymbol(symbol: ts.Symbol): string[];
/**
 * Checks if a symbol is marked as deprecated.
 * @param symbol - The TypeScript symbol.
 * @returns True if the symbol is deprecated; false otherwise.
 */
export declare function isSymbolDeprecated(symbol: ts.Symbol): boolean;
/**
 * ==============================
 * Declaration Utilities
 * ==============================
 */
/**
 * Retrieves all preceding comments (both JSDoc and regular comments) from a given declaration.
 * @param declaration - The AST declaration node.
 * @returns An array of comment strings preceding the declaration.
 */
export declare function getCommentsFromDeclaration(declaration: ts.Declaration): string[];
/**
 * Checks if a declaration is marked as deprecated.
 * @param declaration - The AST declaration node.
 * @returns True if the declaration is deprecated; false otherwise.
 */
export declare function isDeclarationDeprecated(declaration: ts.Declaration): boolean;
/**
 * ==============================
 * JSDoc Utilities
 * ==============================
 */
/**
 * Retrieves all JSDoc tags of a specific kind preceding a node.
 * @param declaration - The AST declaration node.
 * @param tagName - The name of the JSDoc tag to retrieve.
 * @returns An array of JSDoc tags matching the specified name.
 */
export declare function getJsDocTags(declaration: ts.Declaration, tagName: string): ts.JSDocTag[];
